/*
 * qmail-destroy: Remove messages from a qmail queue.
 * Copyright (C) 2015 LinuxMagic
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): William Storey <william@linuxmagic.com>
 *            Shaun A. Johnson <shaun@linuxmagic.com>
 *
 * $Id: queue-remove-util.c 20009 2015-10-16 23:55:53Z william $
 *
 * Utility functions.
 */

#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include "queue-remove-util.h"
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#define MAX_LOG_LENGTH 256

static struct passwd * __qru_getpwnam(const char * const);
static struct group * __qru_getgrnam(const char * const);
static struct passwd * __qru_get_user_group_id(const char * const,
        const char * const);

//! Method to find out if a given directory exists or not.
/*!
 * This also tests inode types so will return FALSE if it exists as
 * a file or a symlink. (POSIX only.. these concepts don't exist in win32...)
 *
 * @param   string  directory path to test
 *
 * @returns bool    exists or not
 *
 * @author Shaun A. Johnson <shaun@linuxmagic.com>
 * @author William Storey <william@linuxmagic.com>
 */
bool qru_directory_exists(const char * const path)
{
    struct stat sb;
    memset(&sb, 0, sizeof(struct stat));

    if (!path || strlen(path) == 0) {
        return false;
    }

    if (stat(path, &sb) != 0) {
        return false;
    }

    if (S_ISDIR(sb.st_mode)) {
        return true;
    }
    return false;
}

//! Method to find out if given file exists or not.
/*!
 * @param   string  file path to test
 *
 * @returns bool    exists or not
 */
bool qru_file_exists(const char *path)
{
    struct stat statbuf;
    int rv = 0;

    if (!path || !strlen(path)) {
        QRUERROR(strerror(EINVAL));
        return false;
    }
    if ( (rv = stat(path,&statbuf)) != 0) {
        // logging to debug because this is part of the test...
        QRUDEBUG(strerror(errno));
        return false;
    }
    // now.. we say 'yes' if it is a regular file or a symlink.
    if (S_ISREG(statbuf.st_mode)) {
        QRUDEBUG("regular file - returning true");
        return true;
    }
    if (S_ISLNK(statbuf.st_mode)) {
        QRUDEBUG("file is a symlink - returning true");
        return true;
    }
    // we should only reach this point if the file is a special type which we shouldn't be dealing with.
    return false;
}

//! Function to open a file in a common 'cross platform' way.
/*!
 * Using this function offers full compatibility
 * on both POSIX as well as windows for binary mode
 * file operations.
 *
 * @param   string      file path to open
 * @param   bool        flag to indicate if the file should be write locked on open
 * @param   integer     mode to open with (man fcntl...)
 *
 * @returns int fd or -1
 */
int qru_open(const char *path, bool lock, int mode)
{
    if (!path) {
        return -1;
    }

    // the file descriptor handle
    int fd = 0;

    struct flock fl = { F_WRLCK, SEEK_SET,0,0,0};
    fl.l_pid = getpid();

    // try to open the file...
    fd = open(path, mode, 0666);
    if (fd == -1) {
        return -1;
    }
    if (lock) {
        if (fcntl(fd,F_SETLKW,&fl) == -1)  {
            close(fd);
            return -1;
        }
    }
    return fd;
}

//! Function to safely close a file handle
/*!
 * This approach offers full compatibility
 * on both POSIX as well as windows systems.
 *
 * @param   int file handle to close
 *
 * @returns bool    success or fail
 */
bool qru_close(int fd)
{
    if (fd == -1) {
        return false;
    }
    if (close(fd) == 0) {
        return true;
    }
    return false;
}

//! Method to determine the current file position for a given handle.
/*!
 * @param   int the file handle to operate on
 *
 * @returns integer the current position from start (in bytes)
 */
int qru_tell(int fd)
{
    if (fd == -1) {
        return -1;
    }
    ssize_t pos = 0;
    pos = lseek(fd, 0, SEEK_CUR);
    if (pos == -1) {
        return -3;
    }
    return pos;
}

//! Method to move the current file position to a given offset
/*!
 * @param   int the file handle to seek in
 * @param   integer the number of bytes to seek
 * @param   integer the starting point to seek from (see SEEK_SET, SEEK_CUR, SEEK_END)
 *
 * @returns bool    success or failure
 */
bool qru_seek(int fd, int offset, int whence)
{
    if (fd == -1) {
        return false;
    }
    if (lseek(fd,offset,whence) == -1) {
        return false;
    }
    return true;
}

//! Method to write a block of data to a given file handle
/*!
 * This is the compatibility equivalent of
 * fwrite and write functions. This method offers full
 * compatibility on both POSIX as well as windows systems.
 *
 * @param   int     the file handle to write to (MUST BE OPEN!)
 * @param   void*   the data to write out (void* since it can be any type)
 * @param   size_t  the size of the data to write out.
 *
 * @returns bool    success or fail.
 */
bool qru_write(int fd, void *data, size_t sz)
{
    void * d_ptr = data;
    size_t bytesLeft = sz;
    size_t block_size = 0;
    ssize_t bytesWritten = 0;

    if (fd == -1 || !data || sz < 1) {
        QRUERROR(strerror(EINVAL));
        return false;
    }

    // we need to do this in blocks...
    while (bytesLeft > 0) {
        // how much do we want to write out?
        // 4096 - block size
        block_size = bytesLeft < 4096 ? bytesLeft : 4096;

        // now try to write it out...
        bytesWritten = write(fd, d_ptr, block_size);
        if (bytesWritten < 0) {
            QRUERROR("write() failure: %s", strerror(errno));
            return false;
        }

        // cast to unsigned - we know it's not negative now. value won't change.
        bytesLeft -= (size_t) bytesWritten;

        d_ptr += bytesWritten;
    }
    return true;
}

//! Method to safely read a block of data from a file.
/*!
 * This will attempt to read up to given amount of data
 * from the file into the given destination memory space.
 * If the data available falls short, then so be it.
 *
 * @param   int     the file handle to read from (MUST BE OPEN!)
 * @param   void*   the destination buffer
 * @param   size_t  the maximum amount to read
 *
 * @returns bool    success or fail
 */
bool qru_read(int fd, void *buffer, size_t sz)
{
    void * d_ptr = buffer;
    size_t bytesLeft = sz;
    size_t block_size = 0;
    ssize_t bytesRead = 0;

    if (fd == -1) {
        QRUERROR("Invalid file handler!");
        return false;
    }
    if (!buffer) {
        QRUERROR("Invalid buffer parameter!");
        return false;
    }
    if (sz < 1) {
        QRUERROR("Invalid read size: %i", sz);
        return false;
    }

    while (bytesLeft > 0) {
        QRUDEBUG("%d bytes remaining to read", bytesLeft);

        // 4096 - block size.
        block_size = bytesLeft < 4096 ? bytesLeft : 4096;

        // try to read a block of data...
        QRUDEBUG("attempting to read %d bytes in one block", block_size);
        bytesRead = read(fd, d_ptr, block_size);
        if (bytesRead < 0) {
            QRUERROR(strerror(errno));
            return false;
        }

        QRUDEBUG("read %d bytes", bytesRead);

        // cast to unsigned - we know it's not negative now. value won't change.
        bytesLeft -= (size_t) bytesRead;
        d_ptr += bytesRead;

        // less than expected means EOF... or some other error...
        if ((size_t) bytesRead < block_size) {
            bytesLeft = 0;
        }
    }

    return true;
}

//! utility function to set the 'run as' identity
/*!
 * @param   string  the username to run as
 *
 * @returns boolean success or fail of change.
 *
 * @author Shaun A. Johnson <shaun@linuxmagic.com>
 * @author William Storey <william@linuxmagic.com>
 *
 * NOTE: if the user running the program is the *SAME* as the
 *       set as then all good, otherwise it requires ROOT privileges
 *       or sticky bit user set on binary.
 */
bool qru_set_run_as(const char * const username)
{
    // sanity check
    if (!username || strlen(username) == 0) {
        return false;
    }

    uid_t uid = 0, ruid = 0;
    struct passwd * pwd = NULL;

    // get the UID
    pwd = __qru_getpwnam(username);
    if (!pwd) {
        QRUERROR("Error retrieving %s UID", username);
        return false;
    }
    ruid = pwd->pw_uid;
    QRUDEBUG("user %s has uid %d", username, ruid);

    // now determine what this program is running as...
    uid = geteuid();
    // and make sure we are within the minimums
    if (uid != 0 && uid != ruid) {
        QRUERROR("This program must be run as 'root' or '%s' user", username);
        free(pwd);
        return false;
    }

    // if this was started as root, then transform...
    if (uid == 0 && ruid != 0) {
        if (seteuid(ruid) != 0) {
            QRUERROR("Error setting effective UID %d: [%s]", ruid,
                strerror(errno));
            free(pwd);
            return false;
        }
    }

    free(pwd);
    return true;
}

//! utility function to set the 'run as' identity: group
/*!
 * @param   string  the group to run as
 *
 * @returns boolean success or fail of change.
 *
 * @author Shaun A. Johnson <shaun@linuxmagic.com>
 * @author William Storey <william@linuxmagic.com>
 *
 * NOTE: if the user running the program is the *SAME* as the
 *       set as then all good, otherwise it requires ROOT privileges
 *       or sticky bit user set on binary.
 */
bool qru_set_run_as_group(const char * const group)
{
    if (!group || strlen(group) == 0) {
        return false;
    }

    gid_t gid = 0, rgid = 0;
    struct group * group_info = NULL;

    group_info = __qru_getgrnam(group);
    if (!group_info) {
        QRUERROR("Error retrieving %s GID", group);
        return false;
    }
    rgid = group_info->gr_gid;
    QRUDEBUG("group %s has gid %d", group, rgid);

    // now determine what this program is running as...
    gid = getegid();

    // and make sure we are within the minimums
    if (gid != 0 && gid != rgid) {
        QRUERROR("This program must be run as 'root' or '%s' group", group);
        free(group_info);
        return false;
    }

    // if this was started as root, then transform...
    if (gid == 0 && rgid != 0) {
        if (setegid(rgid) != 0) {
            QRUERROR("Error setting effective UID %d: [%s]", rgid,
                strerror(errno));
            free(group_info);
            return false;
        }
    }

    free(group_info);
    return true;
}

//! retrieve uid by username. without depending on getpwnam().
/*!
 * @param char* Username we want to look up
 *
 * @returns struct passwd* or NULL if failure
 *   Note only the pw_uid will be set.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * @see __qru_get_user_group_id()
 */
static struct passwd * __qru_getpwnam(const char * const username)
{
    if (!username || strlen(username) == 0) {
        QRUERROR(strerror(EINVAL));
        return NULL;
    }

    return __qru_get_user_group_id("/etc/passwd", username);
}

//! retrieve gid by name. without depending on getgrnam().
/*!
 * @param char* Group name we want to look up
 *
 * @returns struct group * or NULL if failure
 *   Note only the gr_gid will be set.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * @see __qru_get_user_group_id()
 */
static struct group * __qru_getgrnam(const char * const name)
{
    struct passwd * pwd = NULL;
    struct group * group = NULL;

    if (!name || strlen(name) == 0) {
        QRUERROR(strerror(EINVAL));
        return NULL;
    }

    pwd = __qru_get_user_group_id("/etc/group", name);
    if (!pwd) {
        return NULL;
    }

    group = calloc(1, sizeof(struct group));
    if (!group) {
        QRUERROR(strerror(ENOMEM));
        free(pwd);
        return NULL;
    }

    group->gr_gid = pwd->pw_uid;
    free(pwd);
    return group;
}

//! parse id portion from an /etc/passwd or /etc/group file
/*!
 * @param char* The file to look at
 * @param char* Name - username or group
 *
 * @returns struct passwd* or NULL if failure
 *   Note only the pw_uid will be set. Even if we are asked for gid.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * we want to be able to static compile but getpwnam()/getgrnam() are an issue
 * because glibc forces that to be dynamic in some ways.
 *
 * as a result we have this reimplementation where all we do is examine
 * /etc/passwd for the uid and /etc/group for the gid. I use the same
 * function because the parsing is the same for the two files.
 *
 * yes a bit hacky to have one function for both groups and users but I think
 * this better than having two very repeated functions.
 */
static struct passwd * __qru_get_user_group_id(const char * const file,
    const char * const name)
{
    FILE * fh = NULL;
    char * line = NULL;
    // len isn't used but apparently getline() does not like having null.
    size_t len = 0;
    char * line_ptr = NULL;
    int name_len = 0;
    uid_t uid = 0;
    struct passwd * pwd = NULL;

    if (!file || strlen(file) == 0 || !name || strlen(name) == 0) {
        QRUERROR(strerror(EINVAL));
        return NULL;
    }

    fh = fopen(file, "r");
    if (!fh) {
        QRUERROR("fopen(): %s", strerror(errno));
        return NULL;
    }

    while (getline(&line, &len, fh) != -1) {
        // format of line we care about:
        // <name>:<password>:uid:

        line_ptr = strchr(line, ':');
        if (!line_ptr) {
            free(line);
            line = NULL;
            continue;
        }

        // now:
        // <name>:<password>:uid:
        //       ^

        // check if this is the user we care about
        name_len = line_ptr-line;
        if (name_len == 0 || (size_t) name_len != strlen(name)) {
            free(line);
            line = NULL;
            continue;
        }

        if (strncmp(line, name, (size_t) name_len) != 0) {
            free(line);
            line = NULL;
            continue;
        }

        // it is, so get our uid

        line_ptr++;

        // now:
        // <name>:<password>:uid:
        //        ^

        if (*line_ptr == '\0') {
            free(line);
            line = NULL;
            continue;
        }

        line_ptr = strchr(line_ptr, ':');
        if (!line_ptr) {
            free(line);
            line = NULL;
            continue;
        }

        // now:
        // <name>:<password>:uid:
        //                  ^
        line_ptr++;
        if (*line_ptr == '\0') {
            free(line);
            line = NULL;
            continue;
        }

        uid = (unsigned int) atoi(line_ptr);

        free(line);
        line = NULL;
        fclose(fh);

        // we have it.
        pwd = calloc(1, sizeof(struct passwd));
        if (!pwd) {
            QRUERROR(strerror(ENOMEM));
            return NULL;
        }

        pwd->pw_uid = uid;
        return pwd;
    }

    if (fclose(fh) != 0) {
        QRUERROR("fclose(): %s", strerror(errno));
        return NULL;
    }

    QRUDEBUG("not found");
    return NULL;
}

//! method to append a NUMBER to a string as a string.
/*!
 * In the event the destination string is 'null' then
 * a new memory space is created. Otherwise the memory
 * space is resized dynamically.
 *
 * @param   char ** destination memory space pointer
 * @param   integer the number to append
 *
 * @returns boolean success or fail.
 *
 * @author Shaun A. Johnson <shaun@linuxmagic.com>
 */
bool qru_icat(char **dest, long numeric)
{
    // sanity test
    if (!dest) {
        return false;
    }
    char *numbuffer = NULL;
    // stage one.. convert the number to a char buffer
    if (!qru_ltoa(&numbuffer, numeric)) {
        return false;
    }
    // stage 2: strcat this sucker.
    if (!qru_strcat(dest, numbuffer)) {
        free(numbuffer);
        return false;
    }
    free(numbuffer);
    return true;
}

//! Utility function for converting signed numerics to a string
/*!
 * using 'long' type so be aware there may be different maximums in 64 bit
 * environments.  If the destiation memory pointer passed is NULL then a new
 * string is allocated. Be sure to 'free' it when you are done.
 *
 * @param   char**  destination memory pointer.
 * @param   integer the number to convert
 *
 * @returns bool    success or fail
 * @author(s): Shaun A. Johnson <shaun@linuxmagic.com>
 */
bool qru_ltoa(char **dest, long numeric)
{
    char *buffer = 0;
    long num = numeric;
    size_t bufferlen = 2; // even if 0, we still need room for single digit and terminator.

    if (!dest) {
        return false;
    }

    // start by determining the length of buffer required.
    if (num < 0) { // account for '-' sign
        bufferlen++;
        num *= -1;
    }
    while (num >= 10) {
        num /= 10;
        bufferlen++;
    }
    // now we allocate the buffer.
    buffer = calloc(bufferlen,sizeof(char));
    if (!buffer) {
        return false;
    }
    snprintf(buffer,bufferlen,"%ld",numeric);
    if (*dest) {
        free(*dest);
    }
    *dest = buffer;
    return true;
}

//! Method for appending strings to each other.
/*!
 * In the event that the destination string is 'null' then
 * a new memory space is created.  Otherwise, the memory
 * space is resized dynamically.
 *
 * @param   char**  destination memory space pointer
 * @param   const char* source to append
 *
 * @returns bool    success or fail
 * @author(s): Shaun A. Johnson <shaun@linuxmagic.com>
 */
bool qru_strcat(char **dest, const char *src)
{
    if (!dest || !src) {
        return false;
    }

    char *ptr = *dest;
    if (!ptr) {
        ptr = calloc(strlen(src)+1,sizeof(char));
        if (!ptr) {
            return false;
        }
    } else {
        ptr = realloc(*dest,strlen(*dest) + strlen(src) + 1);
        if (!ptr) {
            return false;
        }
    }
    strcat(ptr,src);
    *dest = ptr;
    return true;
}

//! common function for logging information
/*!
 * @param char *format format string for the information to log
 *
 * This function will only log IF the logging
 * hooks have been defined via 'set_logger'
 */
void QRUINFO(const char *format, ...)
{
    // don't move forward if nothing to log or nowhere to log to
    if (!format) {
        return;
    }
    // create our buffer
    char buffer[MAX_LOG_LENGTH];
    memset(buffer, 0, MAX_LOG_LENGTH);
    va_list args;

    // convert the 'printf' style arguments.
    va_start(args, format);
    vsnprintf(buffer, MAX_LOG_LENGTH-1, format, args);
    va_end(args);

    printf("%s\n", buffer);
}

//! common function for logging errors.
/*!
 * @param const char *fn the calling function that caused the error
 * @param const char *file the file that generated the error
 * @param int line number that the error occurred on
 * @param formatted message string printf() as an example
 * @param ... mixed arguments for the formatted message
 *
 * This function will only log IF the logging
 * hooks have been set via set_logger.
 */
void __QRUERROR(const char *fn, const char *file, int line, char *format, ...)
{
    if (!format || !fn) {
        return;
    }

    char buffer[MAX_LOG_LENGTH];
    memset(buffer, 0, MAX_LOG_LENGTH);
    va_list args;

    va_start(args,format);
    vsnprintf(buffer, MAX_LOG_LENGTH-1, format, args);
    va_end(args);

    printf("%s:%d ERROR %s(): %s\n", file, line, fn, buffer);
}

//! common function for logging debug messages.
/*!
 * @param const char *fn the function that caused the debug
 * @param const char *file the file the error occurred in
 * @param int line the line number the error occurred on
 * @param const char *format the formatted message string see printf for example
 *
 * This function will only log IF the logging
 * hooks have been set via set_logger
 */
void __QRUDEBUG(const char *fn, const char *file, int line, const char *format, ...)
{
#ifndef VERBOSEDEBUG
    if (!getenv("VERBOSEDEBUG"))
        return;
#endif

    if (!format || !fn) {
        return;
    }

    char buffer[MAX_LOG_LENGTH];
    memset(buffer, 0, MAX_LOG_LENGTH);
    va_list args;

    va_start(args, format);
    vsnprintf(buffer, MAX_LOG_LENGTH-1, format, args);
    va_end(args);
    printf("%s:%d DEBUG %s(): %s\n", file, line, fn, buffer);
}
